
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  402
#define PPWM 800      // 800 = 20kHz PWM and about 50Hz output

volatile int p1count,p2count,p3count;
volatile int l[NPWM];
volatile uint8_t uf,oen,dbounce,init_done;
volatile int vpwr[3];
volatile float Input[3], Output[3], Setpoint[3];
volatile float errSum[3], lastErr[3],error[3],dErr[3];
volatile float kp, ki, kd;
volatile float ac_output[3],ac_setpoint,pwr[3];
volatile int sst;

void init_vars()
  {
  int i;
  p1count = 0;
  p2count = 134;
  p3count = 268;
  sst = 0;      // slow start counter 0 - 251
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  for (i=0; i < 3; i++)
    {
    pwr[i] = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
    vpwr[i] = 0;      // integer equivalent of pwr, 0 to PPWM
    errSum[i] = 0.0;        // initialise PID
    lastErr[i] = 0.0; 
    }
  dbounce = 0;  // debounce counter for input switch 
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  }

void setup()
  {
  int i;
  float f,t;
  pinMode(7,OUTPUT);
  cbi(PORTH,4);             // drive D7 LOW to disable gate drive outputs from IR2184
  for(i=0; i < NPWM; i++)
    {
    t = 2.0 * 3.14159 * (float)i/(float)NPWM;
    f = (float)PPWM * (1.0 + 0.98*sin(t)) * 0.5;
    l[i] = (int)(f);   
    } 
  t135();
  init_vars();
  pinMode(8,INPUT);       // continuous on/off pin, driven by nano2
  pinMode(13,OUTPUT);      // LED PORTB 7
  pinMode(9,OUTPUT);      // PORTH 6  for 50 Hz sync pulse for CRO trigger etc.

  }

void t135()
  {
  pinMode(46,OUTPUT);
  pinMode(5,OUTPUT);
  pinMode(11,OUTPUT);
  TCCR1A = TCCR3A = TCCR5A = _BV(COM5A1) |  _BV(COM5B1) | _BV(WGM51);
  TCCR1B = TCCR3B = TCCR5B = _BV(WGM53) | _BV(WGM52) | _BV(CS50);
  OCR1A = OCR3A = OCR5A = 1;
  ICR1 = ICR3 = ICR5 = PPWM; 
  TIMSK1 |= (1 << TOIE5); 
  TIMSK3 |= (1 << TOIE5); 
  TIMSK5 |= (1 << TOIE5); 
  }

  
ISR(TIMER1_OVF_vect)        
  {
  OCR1A = ((long)l[p1count] * (long)vpwr[0]) >> 8;
  p1count++;
  if (p1count < 201) sbi(PORTH,6); else cbi(PORTH,6);    // CRO trigger. A square wave of 50Hz
  if (p1count >= NPWM)
    {
      uf = 1;
      p1count = 0;
    }
  }

ISR(TIMER3_OVF_vect)        
  {
  OCR3A = (((long)l[p2count] * (long)vpwr[1]) >> 8);
  p2count++;
  if (p2count >= NPWM)
    p2count = 0;
  }

ISR(TIMER5_OVF_vect)        
  {
  OCR5A = (((long)l[p3count] * (long)vpwr[2]) >> 8);
  p3count++;
  if (p3count >= NPWM)
    p3count = 0;
  }

void do_pid()
{   
  int i;
  double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
  for (i=0; i < 3; i++)
    {
    error[i] = Setpoint[i] - Input[i];
    errSum[i] += (error[i] * timeChange);
    dErr[i] = (error[i] - lastErr[i]) / timeChange;
    Output[i] = kp * error[i] + ki * errSum[i] + kd * dErr[i];
    lastErr[i] = error[i];
    }
}


void loop()
{
  int i;
  
  ac_output[0] = (float)analogRead(0) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
  ac_output[1] = (float)analogRead(1) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
  ac_output[2] = (float)analogRead(2) / 1024.0; // scale 0 - 5V to 0.0 to 1.0   
  
  if (uf == 1)  // only execute this once each 50Hz 
    {
      uf=0;  
      ac_setpoint = 0.55;       // for same Vfb as EG8010. change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;      // slow start. If output is enabled, keep starting up. sst = output pwm% or approx. voltage.
      if (sst > 251)sst = 251;  // stop increasing when sst has got to the end of slow start. when = 251, this enables PID. See below..
      if (oen == 0) sst--;      // if stopping, slow stop
      if (sst <= 0)             // once fully stopped, no more changes to sst needed
        {
          sst = 0;
          cbi(PORTH,4);     // iif sst = 0 then disable gate drive output
        }
      else
         sbi(PORTH,4);      // if sst > 0 then enable gate drive output
         
      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                          // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)              // slow stop? I think of it as de-gauss
          { 
          for (i=0; i < 3; i++)
            {  
            if(ac_output[i] < ac_setpoint)
              pwr[i] += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID.
            else                    // During slow start need to ensure output is tracking setpoint
              pwr[i] -= 1.0/250.0;     // should output get close enough to matter.
            }
          }
        else
          {
          for (i=0; i < 3; i++)
            {
            pwr[i] -= 1.0 /250.0;      // this will execute in slow stop
            }
          }
        }
        
      if (sst == 251)       // once sst = 251, run PID. the above bang-bang control will NOT have executed, because sst = 251
        {                   
        for (i=0; i < 3; i++)
          {  
          Setpoint[i] = ac_setpoint;     
          Input[i] = ac_output[i];
          }  
        do_pid();
        for (i=0; i < 3; i++)
          {
          pwr[i] = pwr[i] + Output[i];
          }
        }
        
      for (i=0; i < 3; i++)
        {      
        if (pwr[i] < 0.01) pwr[i] = 0.01;       // prevent it going too low for some reason
        if (pwr[i] > 0.99) pwr[i] = 0.99;     // and too high      
        vpwr[i] = int(256.0 * pwr[i]);        // scale and convert to int, ready for fast maths in the 20kHz interrupt 
        }
      check_switch_cont(); 
    }
}


void check_switch_cont()
{   // called at 50 Hz
  if ((PINH & 0x20) == 0x20)     
    {
    if (init_done == 0 && sst == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    sbi(PORTB,7); // and light LED on D13
    }

  if (((PINH & 0x20) == 0) )       
    {
    init_done = 0;
    oen = 0;    // slow stop
    cbi(PORTB,7); // and blank LED on D13
    }        
}
